import re
import json

from typing import Any, Dict

from langchain_core.messages import (
    SystemMessage,
    AIMessage,
    FunctionMessage,
    HumanMessage
)


class AgentExecutor:
    def __init__(self, llm, tools: list, system_prompt: str):
        """

        Example tool:
        
        def multiply(**kwargs) -> int:
            # Multiply two integers together.
            print(kwargs)
            n1, n2 = kwargs["n1"], kwargs["n2"]
            return n1 * n2
            
        multiply_tool = {
            "name": "multiply",
            "description": "Multiply two numbers",
            "parameters": {
                "type": "object",
                "properties": {
                    "n1": {
                        "type": "int",
                        "description": "Number one",
                    },
                    "n2": {
                        "type": "int",
                        "description": "Number two",
                    },
                },
                "required": ["n1", "n2"],
            },
            "implementation": multiply,  # Attach the function implementation
        }
        """
        self.llm = llm
        self.tools = tools
        self.system_prompt = system_prompt
        
        tool_description = """Use the function '{name}' to {description}:
{tool_json}"""
        tool_descriptions = [tool_description.format(
            name=tool["name"],
            description=tool["description"],
            tool_json=json.dumps(tool["parameters"])
        ) for tool in tools]
        self.tool_prompt = self._generate_tool_prompt(tool_descriptions)

    def _generate_tool_prompt(self, tool_descriptions: list) -> str:
        tools_text = "\n".join(tool_descriptions)
        return f"""
        {self.system_prompt}
        
        You have access to the following functions:

        {tools_text}

        Don't call the function if there is no need for that

        If you choose to call a function ONLY reply in the following format with no prefix or suffix:

        <function=example_function_name>{{"example_name": "example_value"}}</function>

        Reminder:
        - Function calls MUST follow the specified format, start with <function= and end with </function>
        - Required parameters MUST be specified
        - Only call one function at a time
        - Put the entire function call reply on one line

        If there is no function call available, answer the question **like normal with your current knowledge and do not tell the user about function calls**
        """

    def invoke(self, user_input: str, chat_history: list = None) -> Any:
        if chat_history is None:
            chat_history = []
            # Introduce system prompt to the chat history
            chat_history.append(SystemMessage(content=self.tool_prompt))
        else:
            # if no system prompt available
            if not isinstance(chat_history[0], SystemMessage):
                chat_history = [
                    SystemMessage(content=self.tool_prompt)
                ] + chat_history
        
        # Invoke the LLM
        response = self.llm.invoke(user_input, chat_history=chat_history)
        chat_history.append(HumanMessage(content=user_input))
        
        # Parse and execute the response
        parsed_response = self.parse_tool_response(response)
        if parsed_response:
            response = FunctionMessage(
                content=str(self.execute_tool(parsed_response)),
                name=parsed_response["function"]
            )
        else:
            response = AIMessage(content=response)

        chat_history.append(response)
        
        return {"response": response, "chat_history": chat_history}

    def parse_tool_response(self, response: str) -> Dict[str, Any]:
        function_regex = r"<function=(\w+)>(.*?)</function>"
        match = re.search(function_regex, response)

        if match:
            function_name, args_string = match.groups()
            try:
                args = json.loads(args_string)
                return {
                    "function": function_name,
                    "arguments": args,
                }
            except json.JSONDecodeError as error:
                print(f"Error parsing function arguments: {error}")
                return None
        return None

    def execute_tool(self, parsed_response: Dict[str, Any]) -> Any:
        available_functions = {tool['name']: tool for tool in self.tools}
        
        fun_name = parsed_response["function"]
        arguments = parsed_response["arguments"]
        
        if fun_name in available_functions:
            tool_function = available_functions[fun_name]["implementation"]
            return tool_function(**arguments)
        
        print(f"Function {fun_name} not available.")
        return None


from pydantic import Field
from typing import Any, Dict, Iterator, List, Optional

from langchain_core.callbacks.manager import CallbackManagerForLLMRun
from langchain_core.language_models.llms import LLM
from langchain_core.outputs import GenerationChunk

from transformers import pipeline


class CustomLLM(LLM):
    """A custom chat model that echoes the first `n` characters of the input.

    When contributing an implementation to LangChain, carefully document
    the model including the initialization parameters, include
    an example of how to initialize the model and include any relevant
    links to the underlying models documentation or API.

    Example:

        .. code-block:: python

            model = CustomChatModel(n=2)
            result = model.invoke([HumanMessage(content="hello")])
            result = model.batch([[HumanMessage(content="hello")],
                                 [HumanMessage(content="world")]])Z
    """

    n: int = Field(None, alias='max_output_tokens')
    model_name: str = Field(None, alias='model_name')
    
    pipe: Any = None
    terminators: list = Field([], alias="terminators")

    def __init__(self, model_name: str, n: int = 10**10, **kwargs):
        super(CustomLLM, self).__init__()
    
        self.model_name = model_name
        self.n = n

        model_kwargs = {}

        for arg, value in kwargs.items():
            if 'quantization_config' == arg:
                model_kwargs['quantization_config'] = value
        
        self.pipe = pipeline(
            "text-generation",
            model=model_name,
            model_kwargs=model_kwargs,
            device_map="auto",
        )
        self.terminators = [
            self.pipe.tokenizer.eos_token_id,
            self.pipe.tokenizer.convert_tokens_to_ids("<|eot_id|>")
        ]

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        """Run the LLM on the given input.

        Override this method to implement the LLM logic.

        Args:
            prompt: The prompt to generate from.
            **kwargs: Arbitrary additional keyword arguments. These are usually passed
                to the model provider API call.

        Returns:
            The model output as a string. Actual completions SHOULD NOT include the prompt.
        """
        # if stop is not None:
            # raise ValueError("stop kwargs are not permitted.")
        messages = []
        if "chat_history" in kwargs:
            messages = [{"role": message.type, "content": message.content}
                        for message in kwargs["chat_history"]]

        messages.append({"role": "user", "content": prompt})

        outputs = self.pipe(
            messages,
            max_new_tokens=256,
            eos_token_id=self.terminators,
            do_sample=True,
            temperature=0.6,
            top_p=0.9,
        )

        return outputs[0]["generated_text"][-1]["content"][:self.n]

    def _stream(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[GenerationChunk]:
        """Stream the LLM on the given prompt.

        This method should be overridden by subclasses that support streaming.

        If not implemented, the default behavior of calls to stream will be to
        fallback to the non-streaming version of the model and return
        the output as a single chunk.

        Args:
            prompt: The prompt to generate from.
            stop: Stop words to use when generating. Model output is cut off at the
                first occurrence of any of these substrings.
            run_manager: Callback manager for the run.
            **kwargs: Arbitrary additional keyword arguments. These are usually passed
                to the model provider API call.

        Returns:
            An iterator of GenerationChunks.
        """
        # TODO: implement streaming
        return self._call(prompt)

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Return a dictionary of identifying parameters."""
        return {
            # The model name allows users to specify custom token counting
            # rules in LLM monitoring applications (e.g., in LangSmith users
            # can provide per token pricing for their model and monitor
            # costs for the given LLM.)
            "model_name": self.model_name,
        }

    @property
    def _llm_type(self) -> str:
        """Get the type of language model used by this chat model. Used for logging purposes only."""
        return f"Custom {self.model_name}"


import os
import setuptools


__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def read_requirements():
    """
    Reads the requirements.txt file
    """
    reqs_path = os.path.join(__location__, 'requirements.txt')
    with open(reqs_path, encoding='utf8') as f:
        reqs = [line.strip() for line in f if not line.strip().startswith('#')]

    names = []
    for req in reqs:
        names.append(req)
    return {'install_requires': names}


setuptools.setup(
    name="os-llm-agents",
    version="0.0.1",
    author="Aleksandr Perevalov, Andreas Both",
    author_email="aleksandr.perevalov@htwk-leipzig.de,andreas.both@htwk-leipzig.de",
    description="Python package that helps to build LLM agents based on open-source models from Huggingface Hub.",
    long_description=long_description,
    license="MIT",
    long_description_content_type="text/markdown",
    url="https://github.com/WSE-research/open-source-llm-agents",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    packages=setuptools.find_packages(),
    keywords="llm agents huggingface transformers",
    **read_requirements()
)